#include <oxstd.h>
#include <oxfloat.h>
#include <maximize.h>
#pragma link("maximize.oxo")

// g_n is number of quadrature points in each direction

static decl g_n=100, weights, nodes, 
   response, num_sub, num_rep, total_n, del_dose, del_sub;

// fNodes computes the weights and the nodes for Gauss-Hermite quadrature

fNodes(const weights, const nodes)
{ 
   decl i,j,w, wt, nt, b, mat, mat1 , norm_w, res, ws=zeros(1,g_n*g_n),
   nd=zeros(2,g_n*g_n);
 
   b = sqrt(range(1,g_n-1)./2);
   mat1 = diag(b);
   mat=zeros(g_n,g_n);
   mat[1:g_n-1][0:g_n-2]=mat1;
   mat = mat + mat';
   
   res=eigensym(mat,&nt,&w);
   wt=sqrt(M_PI)*(w[0][].^2);

   for (i=1;i<=g_n;i++)
     for (j=1;j<=g_n;j++)
          {
           nd[0][g_n*(i-1)+j-1]=nt[][i-1];
           nd[1][g_n*(i-1)+j-1]=nt[][j-1];
           ws[][g_n*(i-1)+j-1]=wt[][i-1]*wt[][j-1];
          }
     ws = ws / M_PI;
   nodes[0]=nd;
   weights[0]=ws;
   return 1;
}

// fLikMax computes the approximate likelihood for a given set of parameters
// and returns its value over total sample size.
// It calls fNodes and is used by MaxBFGS

fLikMax(const vP, const adFunc, const avScore, const amHess)
{
     
 decl sigma=vP[4][0]^2,
      beta_1=vP[0:1][0], num_var=2, Sigma_ch=zeros(2,2),
      beta_2=vP[2:3][0], norm_exp, Be_exp,temp_exp, lik_comp,i,j,q, nds;
  
      Sigma_ch[0][0]=vP[5][0];
      Sigma_ch[1][1]=vP[6][0];
      Sigma_ch[1][0]=vP[7][0]; 

      nds = sqrt(2)*Sigma_ch*nodes;

      lik_comp = zeros(num_sub,1); temp_exp = zeros(num_sub,1);
      for (i=0;i<=num_sub-1;i++)
        {
          for (q=1;q<=g_n*g_n;q++)
          { 
        norm_exp = exp(-sumc((response[del_sub[i][]:del_sub[i+1][]-1][0]
                   - del_dose[i][]*beta_1-nds[0][q-1]).^2)/(2*sigma));
        Be_exp = exp(sumc(response[del_sub[i][]:del_sub[i+1][]-1][1])*
                           (del_dose[i][]*beta_2+nds[1][q-1]))/
                (1+exp(del_dose[i][]*beta_2+nds[1][q-1]))^num_rep[i][];
        lik_comp[i][] = lik_comp[i][] + weights[][q-1]*(norm_exp*Be_exp);
      }
      temp_exp[i][]=log(lik_comp[i][]);
     }
      adFunc[0]= sumc(temp_exp)/total_n
                             -0.5*log(2*M_PI*sigma);
  return 1;
}

// fLikMax1 is similar to fLikMax except that it returns the
// approximate likelihood (not divided by sample size) and
// computes the function with respect to the standard deviations
// and correlations not the Cholesky roots of the random effects
// variance-covariance matrix. It is used to obtain standard
// errors of final parameter estimates using numerical
// derivatives.

fLikMax1(const vP, const adFunc, const avScore, const amHess)
{
     
 decl sigma=vP[4][0]^2,
      beta_1=vP[0:1][0], num_var=2, Sigma=zeros(2,2),
      beta_2=vP[2:3][0], norm_exp, Be_exp,temp_exp, lik_comp,i,j,q, nds;
  
      Sigma[0][0]=vP[5][0]^2;
      Sigma[1][1]=vP[6][0]^2;
      Sigma[1][0]=Sigma[0][1]=
	     vP[7][0]*vP[5][0]*vP[6][0]; 

      nds = sqrt(2)*choleski(Sigma)*nodes;

      lik_comp = zeros(num_sub,1); temp_exp = zeros(num_sub,1);
      for (i=0;i<=num_sub-1;i++)
        {
          for (q=1;q<=g_n*g_n;q++)
          { 
        norm_exp = exp(-sumc((response[del_sub[i][]:del_sub[i+1][]-1][0]
                   - del_dose[i][]*beta_1-nds[0][q-1]).^2)/(2*sigma));
        Be_exp = exp(sumc(response[del_sub[i][]:del_sub[i+1][]-1][1])*
                           (del_dose[i][]*beta_2+nds[1][q-1]))/
                (1+exp(del_dose[i][]*beta_2+nds[1][q-1]))^num_rep[i][];
        lik_comp[i][] = lik_comp[i][] + weights[][q-1]*(norm_exp*Be_exp);
      }
      temp_exp[i][]=log(lik_comp[i][]);
     }
      adFunc[0]= sumc(temp_exp)
                             -0.5*total_n*log(2*M_PI*sigma);
  return 1;
}

main()
{  decl n, vec, dfunc, nt=zeros(1,g_n), wt=zeros(1,g_n), amHess=unit(8),
        delta=0.0001, avScore=zeros(8,1),avScore1=avScore, amHess1=unit(8),
        Succ1, Succ2, Succ3, Succ4, time, litter, dose,
        r, result, var_est,
		etalon,bool,data,i, Sigma=zeros(2,2), Sigma_ch; 
// vec is a vector of parameter estimates in the
// following order: beta_1, beta_2, sigma, sigma_11
// sigma_22, sigma_12.
// 

   fNodes(&weights,&nodes);

// set initial values for the parameter estimates.
// Note that vec[4][0] contains the signed std. deviation of sigma.
// Also vec[5][0], vec[6][0] and vec[7][0] are the (1,1), (2,2) and
// (2,1) elements of the Choleski decomposition of Sigma.

   vec=zeros(8,1);
   vec[0][0] = 0.94966;
   vec[1][0] = -0.090854;
   vec[2][0] = -2.9490;
   vec[3][0] = 1.1792;
   vec[4][0] = 0.111243;


    Sigma[0][0]=1;
    Sigma[0][1]=Sigma[1][0]=0;
    Sigma[1][1]=1;

    Sigma_ch=choleski(Sigma);
   
    vec[5][0]=Sigma_ch[0][0];
    vec[6][0]=Sigma_ch[1][1];
    vec[7][0]=Sigma_ch[1][0];

    data = loadmat ("EG.mat");    // read in the data

    litter = data[][1];
    dose = 1 ~ (data[][2]/1000);
    response = data[][4:5];
    
    i=0; del_sub=0;num_rep=0;
            /* creating a vector with the number of subjects in each litter */
    do
     { etalon=litter[i][]*ones(rows(litter));
       bool=(etalon .== litter);
       i = i + sumc(bool); del_sub=del_sub |i;
       num_rep = num_rep | sumc(bool);
     }
     while (i < rows(litter));
     num_rep=num_rep[1:][];
     num_sub=rows(num_rep);
     total_n=del_sub[rows(del_sub)-1][0];

     del_dose = dose[del_sub[:num_sub-1][]][];

// Maximization of the approximate likelihood is performed
// using the MaxBFGS function in Ox.
	 
   MaxControl(100,1);
   MaxControlEps(1e-4,1e-3);
   MaxBFGS(fLikMax,&vec,&dfunc,&amHess,1); 
   
   Sigma_ch=zeros(2,2);
   Sigma_ch[0][0]=vec[5][0];
   Sigma_ch[1][1]=vec[6][0];
   Sigma_ch[1][0]=vec[7][0];

   Sigma=Sigma_ch*Sigma_ch';

   vec[4][0]=sqrt(vec[4][0]*vec[4][0]);
   vec[5][0]=sqrt(Sigma[0][0]);
   vec[6][0]=sqrt(Sigma[1][1]);
   vec[7][0]=
   Sigma[1][0]/sqrt(Sigma[0][0]*Sigma[1][1]);
   
   print("The order of the parameters is as follows:","\n",
   "beta_10,beta_11,beta_20,beta_21,","\n",
   "sigma,sigma_b1,sigma_b2,rho_b","\n");
   
   print("param vector =", vec);  

// The standard errors of the final parameter estimates are
// obtained using numerical derivatives.
   
   Succ2 = Num2Derivative(fLikMax1,vec,&amHess); 
   var_est = - invert(amHess);
   r = rows(var_est);
   result = diagonalize(var_est)*ones(r,1);
   result = sqrt(fabs((result .> 0) .* result));
   print("std.errors=",result);
 }





