#include <oxstd.h>
#include <oxfloat.h>

// fRan_MVN generates multivariate normal variates

fRan_MVN(const mu, const Sigma, const num_sim)
 {
   decl gen_data=zeros(columns(Sigma),num_sim), i;
   
   for (i=1;i<=num_sim;i++)
     gen_data[][i-1]=choleski(Sigma)*rann(columns(Sigma),1)+mu;
   
   return(gen_data);
}

// sum_b, sum_bg, sum_by and sum_bgy compute sums used repeatedly
// by the Var_est function. 

sum_b (const beta, const total_n, const response, const dose)
{
  decl i, s2 = zeros(2,2);

  for (i=0; i<=total_n-1; i++)
     s2 = s2 + exp(dose[i][]*beta')/((1+exp(dose[i][]*beta'))^2).*
                                 (dose[i][]'*dose[i][]);
  
  return s2;
}

sum_bg (const beta, const num_sub, const num_rep, 
        const response, const del_dose, const gdata)
{
  decl i, s2 = zeros(2,2);

  for (i=0; i<=num_sub-1; i++)
     s2 = s2 + sumc(exp(del_dose[i][]*beta'+gdata[][i])
                      ./((1+exp(del_dose[i][]*beta'+gdata[][i])).^2)).*
                    (num_rep[i][].*(del_dose[i][]'*del_dose[i][]));
  
  return s2;
}


sum_by (const beta, const total_n, const response, const dose)
{
  decl i, s = zeros(2,1);

  for (i=0; i<=total_n-1; i++)
     s = s + response[i][].*(dose[i][]') - 
             exp(dose[i][]*beta')/(1+exp(dose[i][]*beta')).*(dose[i][]');
  
  return s;
}


sum_bgy (const beta, const num_sub, const num_rep, 
        const response, const del_dose, const gdata, const ss, const mean_y2)
{
  decl i, s2 = zeros(2,1);

  for (i=0; i<=num_sub-1; i++)
     s2 = s2 + ss*mean_y2[i][].*del_dose[i][]'-
               sumc(exp(del_dose[i][]*beta'+gdata[][i])
                      ./(1+exp(del_dose[i][]*beta'+gdata[][i]))).*
                    (num_rep[i][].*(del_dose[i][])');
  
  return s2;
}

// Var_est computes the sandwich variance estimate used in the
// criteria for sample size increase and estimate the observed
// information matrix.

Var_est (const response, const del_dose, const gen_data1, const gen_data2,
           const num_sub, const num_rep, const total_n, const ss,
           const beta, const Sigma_c, const sigma_hat, 
           const sum_x, const sum_xy, const mean_y1, mean_y2,
           const info_matrix)
{  
   decl sum_x2, sum_ber1, sum_bx, sum_bb, temp,
        sum_b1, sum_sqb1, sum_sqb2, sum_b1b2, qu_m=zeros(8,8), 
        v_hat=zeros(8,8), v1_hat=v_hat, exp_est, corr_fac,
        L_der=zeros(8,ss), det, i, j, var_est, var1_est,
        lik_vec, var_loglik, num_var=2, 
        matr1=zeros(2,2), matr2=matr1, matr3=matr1,
        sigma_b1,sigma_b2,rho_b,der_b1,der_b2,der_b,rand_var,k;
    
   sum_x2 = sum_bg(beta[1][],num_sub,num_rep, response[][1],del_dose,
                 gen_data2);
   sum_ber1 = sumc(num_rep.*(sumc(gen_data1))'.*del_dose);
   sum_sqb1 = sumr(sumc(gen_data1[][].^2));
   sum_sqb2 = sumr(sumc(gen_data2[][].^2));
   sum_b1b2 = sumr(sumc(gen_data1[][].*gen_data2[][]));

   matr1[0][0]=1;
   matr2[1][1]=1;
   matr3[1][0]=matr3[0][1]=1; 

   sigma_b1 = sqrt(Sigma_c[0][0]);
   sigma_b2 = sqrt(Sigma_c[1][1]);
   rho_b = Sigma_c[0][1]/(sigma_b1*sigma_b2);
   der_b1 = 2*sigma_b1*matr1+rho_b*sigma_b2*matr3;
   der_b2 = 2*sigma_b2*matr2+rho_b*sigma_b1*matr3;
   der_b = sigma_b1*sigma_b2*matr3; 
   rand_var = invert(Sigma_c);

   qu_m[0:1][0:1] = - sum_x/sigma_hat;
   qu_m[2:3][2:3] = - sum_x2/ss;

   qu_m[4][0:1] = -2*sqrt(sigma_hat)*(sumc(mean_y1.*del_dose)
                                 - beta[0][] * sum_x
                   - sum_ber1/ss)/(sigma_hat^2);
   qu_m[0:1][4] = qu_m[4][0:1]';
   qu_m[4][4] = total_n / sigma_hat -
             3*(sumc(response[][0].^2) + (sumc(num_rep.*sumr((gen_data1'+
                     (del_dose*beta[0][]')).^2)))/ss
                - 2*(sumc(mean_y1.*sumr(gen_data1'+
                    (del_dose*beta[0][]'))))/ss)/(sigma_hat^2);  

   qu_m[5][5]= - 0.5*num_sub*trace(rand_var*der_b1*rand_var*der_b1);
   qu_m[6][6]= - 0.5*num_sub*trace(rand_var*der_b2*rand_var*der_b2);
   qu_m[7][7]= - 0.5*num_sub*trace(rand_var*der_b*rand_var*der_b);
   qu_m[5][6]=qu_m[6][5] = -0.5*num_sub*trace(rand_var*der_b1*rand_var*der_b2);
   qu_m[5][7]=qu_m[7][5] = -0.5*num_sub*trace(rand_var*der_b1*rand_var*der_b);
   qu_m[6][7]=qu_m[7][6] = -0.5*num_sub*trace(rand_var*der_b2*rand_var*der_b);

   sum_bx=zeros(2,ss);
   for (i=0;i<=num_sub-1;i++)
      sum_bx=sum_bx+num_rep[i][].*(gen_data1[][i]'|gen_data1[][i]')
                                                  .*del_dose[i][]';
   sum_bb=zeros(2,ss);
   for (i=0;i<=num_sub-1;i++)
      { temp=exp(del_dose[i][]*beta[1][]'+ gen_data2[][i]);
      sum_bb=sum_bb+num_rep[i][].*((temp./(1+temp))'|(temp./(1+temp))').*
                                             del_dose[i][]';
      }

   L_der[0:1][] = (sum_xy - sum_x*beta[0][]' - sum_bx)/sigma_hat;
   L_der[2:3][] = (sumc(num_rep.*mean_y2.*del_dose))' - sum_bb;
   L_der[4][] = - total_n / sqrt(sigma_hat) + (sumc(response[][0].^2)
                + sumc(num_rep.*(del_dose*beta[0][]' + gen_data1').^2)
                - 2*sumc(mean_y1.*(gen_data1'+del_dose*beta[0][]')))/
                                    ((sqrt(sigma_hat)^3)); 
   for (k=0;k<ss;k++)
    {
      sum_bb = zeros(2,2);
      sum_bb[0][0] = sumr(gen_data1[k][].^2);
      sum_bb[1][1] = sumr(gen_data2[k][].^2);
      sum_bb[0][1] = sum_bb[1][0] = sumr(gen_data1[k][].*gen_data2[k][]);
  L_der[5][k] = 0.5*trace(rand_var*(sum_bb - num_sub*Sigma_c)*rand_var*der_b1);
  L_der[6][k] = 0.5*trace(rand_var*(sum_bb - num_sub*Sigma_c)*rand_var*der_b2);
  L_der[7][k] = 0.5*trace(rand_var*(sum_bb - num_sub*Sigma_c)*rand_var*der_b);
    }
 
   for (i=0;i<=7;i++)
    { v_hat[i][i]=sumr(L_der[i][].^2);
      for (j=i+1;j<=7;j++)
        v_hat[i][j]=v_hat[j][i]=sumr(L_der[i][].*L_der[j][]);
    }

   exp_est = sumr(L_der[][])/ss;
   corr_fac = exp_est*(exp_est');
   v_hat=v_hat/ss;
   v1_hat=v_hat-corr_fac;
   var1_est= -qu_m -v1_hat;
   info_matrix[0] = var1_est;
   v_hat=v_hat/ss;
   var_est=qu_m*invert(v_hat)*qu_m;
   return(var_est);
}

main() 
{

// ss is simulation sample size
// ss_inc is the fraction used in sample size increase
// delta and eps are precisions used for convergence

    decl  data, temp, temp1, mean_hat, sigma_hat=1, log_tau1, log_tau2,log_tau,
             u, i, j, r, ss=100, ss_inc=3, Be_corr=0.005,
             mu_c=<0;0>, Sigma_c=<1,0;0,1>, 
             num_sub,
             num_var=2, num_rep, e_hat, p_hat, w, beta=zeros(2,2), old_beta,
             c_s, c_f, gen_data1, gen_data2, cindex, time, eps=0.003, 
	     eps1=0.0001,
             delta=0.001, ell, three, last,
             beta_prev, Sigma_prev, sigma_prev, iter, ksai, v_star, qu_m,
             vers1, vers2,var_est,result,
             litter,dose,response,del_sub,total_n, 
             etalon, bool, info_matrix, temp_info,
             sum_x, sum_xy, mean_mat, mean_y1, mean_y2, del_dose;
   
    data = loadmat ("EG.mat"); /* loading of the data */
    litter = data[][1];
    dose = 1 ~ (data[][2]/1000);
    response = data[][4:5];
    
    ranseed(691029);
    i=0; del_sub=0;num_rep=0;
            /* creating a vector with the number of subjects in each litter */
    do
     { etalon=litter[i][]*ones(rows(litter));
       bool=(etalon .== litter);
       i = i + sumc(bool); del_sub=del_sub |i;
       num_rep = num_rep | sumc(bool);
     }
     while (i < rows(litter));
     num_rep=num_rep[1:][];
     num_sub=rows(num_rep);
     total_n=del_sub[rows(del_sub)-1][0];

     log_tau = zeros(num_sub,1);
    
     time=timer();

// Compute 2 sums that will be needed repeatedly in the program

    sum_x=zeros(2,2); sum_xy=zeros(2,1);
    for (i=1; i<=total_n; i++)
      { sum_x=sum_x + dose[i-1][]'*dose[i-1][];
        sum_xy= sum_xy + response[i-1][0]*dose[i-1][]';
      }

// Now comes the rejection sampler.
    
    beta[0][] = (invert(sum_x)*sum_xy)';
    sigma_hat = meanc((response[][0]-dose[][]*beta[0][]').^2);  

    beta[1][]=<0,0>;
    do
    {
    old_beta = beta[1][];
    beta[1][] = old_beta + (invert(sum_b(old_beta,total_n, response[][1],dose))
                         * sum_by(old_beta,total_n, response[][1],dose))';  
   } while (any(fabs((old_beta-beta[1][])./old_beta) .> eps1));   

    three = 0;
    ksai = zeros(8,1);
    v_star = zeros(8,8);
    qu_m = zeros(8,8);
    iter = 0;
    info_matrix = new array[100];

    del_dose = dose[del_sub[:num_sub-1][]][];
  
do
   { 
     print("iteration = ", iter, "\n");
     print("simulation sample size = ", ss, "\n");
     beta_prev=beta;
     Sigma_prev=Sigma_c;
     sigma_prev=sigma_hat;
     gen_data1 = zeros(ss,num_sub);
     gen_data2 = zeros(ss,num_sub);
     for (i=0; i <= num_sub-1; i++)
        {  
          log_tau1 = - num_rep[i][]/2 * log(M_2PI * sigma_hat) 
                    - num_rep[i][]/(2 * sigma_hat) * 
                      varc(response[del_sub[i][]:del_sub[i+1][]-1][0]);
           
          mean_hat = meanc(response[del_sub[i][]:del_sub[i+1][]-1][1]);
          if (mean_hat == 0) mean_hat=mean_hat+Be_corr;
          if (mean_hat == 1) mean_hat=mean_hat-Be_corr;
          log_tau2 = num_rep[i][] * (log(1-mean_hat) + 
                     mean_hat*log((mean_hat/(1-mean_hat))));
  
          log_tau[i][] = log_tau1 + log_tau2; 

          r=0; c_s=0; c_f=0;
          while (r < ss)
           { u = fRan_MVN(mu_c,Sigma_c,1);
             w = ranu(1,1);
             temp1 = - num_rep[i][]/2*log(M_2PI*sigma_hat) 
                     - 1/(2*sigma_hat)*
             sumc((response[del_sub[i][]:del_sub[i+1][]-1][0] -
                       beta[0][]*(del_dose[i][])' - u[0][]).^2)
           + sumc(response[del_sub[i][]:del_sub[i+1][]-1][1])*
                (beta[1][]*del_dose[i][]'+u[1][])
               - num_rep[i][]*log(1+exp(beta[1][]*del_dose[i][]'+u[1][]))
           - log_tau[i][]; 
         if (temp1 > 1)  { print("OOPs", "u=",u,"temp1=",temp1);}
         if (log(w) <= temp1) { gen_data1[r][i] = u[0][];
                                gen_data2[r][i] = u[1][];
                                r=r+1;
                                c_s=c_s+1;}
         else c_f = c_f + 1;
       }
  }
    // Now comes the maximization step

    // Maximizing the normal component
    
    // First create a special matrix to find a vector of means by subject 

   mean_mat = zeros(num_sub, total_n);
   for (i=0; i<=num_sub-1; i++)
      { mean_mat[i][del_sub[i]:del_sub[i+1]-1] = 1; }
   mean_y1 = mean_mat * response[][0];
   mean_y2 = mean_mat * response[][1];


   beta[0][] = (sumc(del_dose
       .*(mean_y1-num_rep.*(meanc(gen_data1))')))*invert(sum_x);

   sigma_hat=meanc(response[][0].^2) + (sumc(num_rep.*sumr((gen_data1'+
      (del_dose*beta[0][]')).^2)))/(total_n*ss)
       - 2*(sumc(mean_y1.*sumr(gen_data1'+
      (del_dose*beta[0][]'))))/(total_n*ss);

   // Maximizing the Bernoulli component

    do
    {
    old_beta = beta[1][];
    beta[1][] = old_beta + 
                (invert(sum_bg(old_beta,num_sub, num_rep,
                 response[][1],del_dose,gen_data2))
               * sum_bgy(old_beta,num_sub,num_rep, response[][1],del_dose,
                 gen_data2, ss, mean_y2))';  
   } while (any(fabs((old_beta-beta[1][])./old_beta) .> eps1));   
 
   // Maximizing with respect to Sigma

   Sigma_c[0][0] = meanc(meanr(gen_data1[][] .^ 2));
   Sigma_c[1][1] = meanc(meanr(gen_data2[][] .^ 2));
   Sigma_c[0][1] = meanc(meanr(gen_data1[][] .* gen_data2[][]));
   Sigma_c[1][0] = Sigma_c[0][1];

   iter=iter+1; 

// Assessment whether simulation sample size needs to be increased
// and whether convergence has occured.
   
   ksai[0:1][] = (beta_prev[0][] - beta[0][])';
   ksai[2:3][] = (beta_prev[1][] - beta[1][])'; 
   ksai[4][] = sqrt(sigma_hat) - sqrt(sigma_prev);
   ksai[5][] = sqrt(Sigma_c[0][0]) - sqrt(Sigma_prev[0][0]);
   ksai[6][] = sqrt(Sigma_c[1][1]) - sqrt(Sigma_prev[1][1]);
   ksai[7][] = Sigma_c[0][1]/sqrt(Sigma_c[0][0]*Sigma_c[1][1]) 
             - Sigma_prev[0][1]/sqrt(Sigma_prev[0][0]*Sigma_prev[1][1]);
			 
   ell = ksai'*
   Var_est(response,del_dose,gen_data1,gen_data2,
           num_sub,num_rep,total_n,ss,
   beta,Sigma_c,sigma_hat,sum_x,sum_xy,mean_y1,mean_y2,
                              &temp_info)*
   ksai;
   info_matrix[iter-1] = temp_info;

   if (ell <= quanchi(0.75,8)) ss=ss+idiv(ss,ss_inc);

   last = !(any(fabs(beta_prev-beta)./(fabs(beta)+delta) .> eps) ||
          (fabs(sqrt(Sigma_c[0][0])-sqrt(Sigma_prev[0][0]))
             /(sqrt(Sigma_c[0][0])+delta)> eps) ||
          (fabs(sqrt(Sigma_c[1][1])-sqrt(Sigma_prev[1][1]))
             /(sqrt(Sigma_c[1][1])+delta)> eps) || 
          (fabs(Sigma_c[0][1]/sqrt(Sigma_c[0][0]*Sigma_c[1][1])
             -(Sigma_prev[0][1]/sqrt(Sigma_prev[0][0]*Sigma_prev[1][1])))
     /(fabs(Sigma_c[0][1]/sqrt(Sigma_c[0][0]*Sigma_c[1][1]))+delta)> eps) || 
   fabs(sqrt(sigma_hat) - sqrt(sigma_prev))/(sqrt(sigma_hat)+delta) > eps);  
   three = last*(three + last);     

   } while (three < 3);

   ksai[0:1][] = (beta[0][])';
   ksai[2:3][] = (beta[1][])'; 
   ksai[4][] = sqrt(sigma_hat);
   ksai[5][] = sqrt(Sigma_c[0][0]);
   ksai[6][] = sqrt(Sigma_c[1][1]);
   ksai[7][] = Sigma_c[0][1]/sqrt(Sigma_c[0][0]*Sigma_c[1][1]); 

   print("The order of the parameters is as follows:","\n",
   "beta_10,beta_11,beta_20,beta_21,","\n",
   "sigma,sigma_b1,sigma_b2,rho_b","\n");
   
   print("Parameters:",ksai); 

// Computation of final approximation of the information matrix.
   
   vers1 = zeros(8,8); vers2 = zeros(8,8);
   j = 0;
   for(i=(iter-1);i>10;i--)
    {
      vers1 = vers1 + info_matrix[i];
      var_est = invert(vers1/(iter-i));
      r = rows(var_est);
      result = diagonalize(var_est)*ones(r,1);
      result = sqrt(fabs((result .> 0) .* result));
	  j = j+1; if (j==3)
	  {print("std.errors=",result);}
    }
 }










