#include <oxstd.h>
#include <oxfloat.h>
#include <maximize.h>
#pragma link("maximize.oxo")

static decl g_n=100, weights,
                    nodes, 
            response, num_sub, num_rep, total_n, del_dose, del_sub;

// fNodes computes the weights and the nodes for Gauss-Hermite quadrature

fNodes(const weights, const nodes)
{ 
   decl i,j,w, wt, nt, b, mat, mat1 , norm_w, res, ws=zeros(1,g_n*g_n),
   nd=zeros(2,g_n*g_n);
 
   b = sqrt(range(1,g_n-1)./2);
   mat1 = diag(b);
   mat=zeros(g_n,g_n);
   mat[1:g_n-1][0:g_n-2]=mat1;
   mat = mat + mat';
   
   res=eigensym(mat,&nt,&w);
   wt=sqrt(M_PI)*(w[0][].^2);

   for (i=1;i<=g_n;i++)
     for (j=1;j<=g_n;j++)
          {
           nd[0][g_n*(i-1)+j-1]=nt[][i-1];
           nd[1][g_n*(i-1)+j-1]=nt[][j-1];
           ws[][g_n*(i-1)+j-1]=wt[][i-1]*wt[][j-1];
          }
     ws = ws / M_PI;
   nodes[0]=nd;
   weights[0]=ws;
   return 1;
}

// fLikMax computes the approximate likelihood for a given set of parameters
// It calls fNodes and is used by MaxBFGS

fLikMax(const vP, const adFunc, const avScore, const amHess)
{
     
 decl sigma=vP[5][0]^2, gamma=vP[2][0], temp_resp,
      beta_1=vP[0:1][0], num_var=2, Sigma_ch=zeros(2,2),
      beta_2=vP[3:4][0], norm_exp, Be_exp,temp_exp, lik_comp,i,j,q, nds;
  
      Sigma_ch[0][0]=vP[6][0];
      Sigma_ch[1][1]=vP[7][0];
      Sigma_ch[1][0]=vP[8][0]; 

      nds = sqrt(2)*Sigma_ch*nodes;

      temp_resp=2*response[][1]-1;
      lik_comp = zeros(num_sub,1); temp_exp = zeros(num_sub,1);
      for (i=0;i<=num_sub-1;i++)
        {
          for (q=1;q<=g_n*g_n;q++)
          { 
        norm_exp = exp(-sumc((response[del_sub[i][]:del_sub[i+1][]-1][0]
                   - del_dose[i][]*beta_1-gamma*
                   temp_resp[del_sub[i][]:del_sub[i+1][]-1][]     
                   -nds[0][q-1]).^2)/(2*sigma));
        Be_exp = exp(sumc(response[del_sub[i][]:del_sub[i+1][]-1][1])*
                           (del_dose[i][]*beta_2+nds[1][q-1]))/
                (1+exp(del_dose[i][]*beta_2+nds[1][q-1]))^num_rep[i][];
        lik_comp[i][] = lik_comp[i][] + weights[][q-1]*(norm_exp*Be_exp);
      }
      temp_exp[i][]=log(lik_comp[i][]);
     }
      adFunc[0]= sumc(temp_exp)/total_n-0.5*log(2*M_PI*sigma);
  return 1;
}

fLikMax1(const vP, const adFunc, const avScore, const amHess)
{
     
 decl sigma=vP[5][0]^2, gamma=vP[2][0], temp_resp,
      beta_1=vP[0:1][0], num_var=2, Sigma=zeros(2,2),
      beta_2=vP[3:4][0], norm_exp, Be_exp,temp_exp, lik_comp,i,j,q, nds;
  
      Sigma[0][0]=vP[6][0]^2;
      Sigma[1][1]=vP[7][0]^2;
      Sigma[1][0]=Sigma[0][1]=vP[8][0]*vP[6][0]*vP[7][0]; 

      nds = sqrt(2)*choleski(Sigma)*nodes;

      temp_resp=2*response[][1]-1;
      lik_comp = zeros(num_sub,1); temp_exp = zeros(num_sub,1);
      for (i=0;i<=num_sub-1;i++)
        {
          for (q=1;q<=g_n*g_n;q++)
          { 
        norm_exp = exp(-sumc((response[del_sub[i][]:del_sub[i+1][]-1][0]
                   - del_dose[i][]*beta_1-gamma*
                   temp_resp[del_sub[i][]:del_sub[i+1][]-1][]     
                   -nds[0][q-1]).^2)/(2*sigma));
        Be_exp = exp(sumc(response[del_sub[i][]:del_sub[i+1][]-1][1])*
                           (del_dose[i][]*beta_2+nds[1][q-1]))/
                (1+exp(del_dose[i][]*beta_2+nds[1][q-1]))^num_rep[i][];
        lik_comp[i][] = lik_comp[i][] + weights[][q-1]*(norm_exp*Be_exp);
      }
      temp_exp[i][]=log(lik_comp[i][]);
     }
      adFunc[0]= sumc(temp_exp)-0.5*total_n*log(2*M_PI*sigma);
  return 1;
}

main()
{  decl n, vec, dfunc, nt=zeros(1,g_n), wt=zeros(1,g_n), amHess=unit(9),
        delta=0.0001, avScore=zeros(9,1),avScore1=avScore, amHess1=unit(9),
        Succ2, time, litter, dose, result, score, tmp,
        etalon,bool,data,i, Sigma=zeros(2,2), Sigma_ch,
		var_est,r; 
                   // vec is a vector of parameter estimates in the
                   // following order: beta_1, beta_2, sigma, sigma_11
                   // sigma_22, sigma_12
  
   fNodes(&weights,&nodes);

   vec=zeros(9,1);
   vec[0][0] = 0.94966;
   vec[1][0] = -0.090854;
   vec[2][0] = 0;
   vec[3][0] = -2.9490;
   vec[4][0] = 1.1792;
   vec[5][0] = sqrt(0.012375);
      // set initial values for the parameter estimates

   vec[6][0]=1;
   vec[7][0]=1;
   vec[8][0]=0;

   data = loadmat ("EG.mat");    // read in the data

   litter = data[][1];
   dose = 1 ~ (data[][2]/1000);
   response = data[][4:5];
    
   i=0; del_sub=0;num_rep=0;
            /* creating a vector with the number of subjects in each litter */
    do
     { etalon=litter[i][]*ones(rows(litter));
       bool=(etalon .== litter);
       i = i + sumc(bool); del_sub=del_sub |i;
       num_rep = num_rep | sumc(bool);
     }
     while (i < rows(litter));
     num_rep=num_rep[1:][];
     num_sub=rows(num_rep);
     total_n=del_sub[rows(del_sub)-1][0];

     del_dose = dose[del_sub[:num_sub-1][]][];

   time=timer();

   MaxControl(100,1);
   MaxControlEps(1e-4,1e-3);
   MaxBFGS(fLikMax,&vec,&dfunc,&amHess,1); 

   Sigma_ch=zeros(2,2);
   Sigma_ch[0][0]=vec[6][0];
   Sigma_ch[1][1]=vec[7][0];
   Sigma_ch[1][0]=vec[8][0];
 
   Sigma=Sigma_ch*Sigma_ch';

   vec[5][0]=sqrt(vec[5][0]*vec[5][0]);
   vec[6][0]=sqrt(Sigma[0][0]);
   vec[7][0]=sqrt(Sigma[1][1]);
   vec[8][0]=Sigma[1][0]/(vec[6][0]*vec[7][0]);

   print("The order of the parameters is as follows:","\n",
   "beta_10,beta_11,gamma,","\n",
   "beta_20,beta_21,","\n",
   "sigma,sigma_b1,sigma_b2,rho_b","\n");
   
   print("param vector =", vec);  
   
   Num2Derivative(fLikMax1,vec,&amHess); 
   var_est = - invert(amHess);
   r = rows(var_est);
   result = diagonalize(var_est)*ones(r,1);
   result = sqrt(fabs((result .> 0) .* result));
   print("std.errors=",result);

 }





